/// A command to be executed by the VM
abstract class Command<T> {
  final T data;

  const Command(this.data);
}

/// A single value stored in the Flight VM
abstract class FlightValue<T> {
  final T value;

  const FlightValue(this.value);

  @override
  bool operator ==(Object other) =>
      other is FlightValue && other.value == this.value;

  @override
  int get hashCode => value.hashCode;

  @override
  String toString() {
    return '$name,$value';
  }

  String get name;
}

/// The Machine interface specifies what to do when a command is called
abstract class VirtualMachine {
  dynamic call(Command command);
}

/// A 'function' in Flight is just a list of commands to execute
typedef FlightFunction = Iterable<Command>;

/// A dartfunction called from the flight vm is passed the vm interface instance
typedef DartFunction = void Function(VirtualMachine);

/// A stack frame, with references to its parent frames so that closures can be implemented
/// Not really an interface - but it has no dependencies and they're pretty integral to closures
class Frame {
  final Frame? parent;
  final Map<FlightValue, FlightValue> data = {};

  Frame({this.parent});

  /// Sets a given key-value in this frame.
  /// If shadows is true (default) always sets in this frame
  /// If false, this will find the nearest ancestor that has the key and override it
  /// If no ancestor has the key, it is set locally
  void set(FlightValue key, FlightValue value, {bool shadows = true}) {
    if (shadows) {
      data[key] = value;
    } else {
      final outerFrame = _frameForKey(key) ?? this;
      outerFrame.set(key, value, shadows: true);
    }
  }

  /// Get the given key from the frame
  FlightValue get(FlightValue key) {
    final frame = _frameForKey(key);
    if (frame == null) {
      throw StateError("Variable does not exist with name ${key.toString()}");
    }
    return frame.data[key]!;
  }

  /// Gets the nearest frame that contains the given key
  /// Returns null if no frames have the key in them
  Frame? _frameForKey(FlightValue key) {
    var current = this;

    while (true) {
      if (current.data.keys.contains(key)) {
        return current;
      }
      if (current.parent == null) {
        return null;
      }
      current = current.parent!;
    }
  }
}
