import 'commands.dart';
import 'interfaces.dart';
import 'values.dart';

class MachineImpl implements VirtualMachine {
  final List<FlightValue> stack = [];

  final List<Frame> frameStack = [Frame()];

  @override
  dynamic call(Command command) {
    switch (command.runtimeType) {
      case Push:
        return _push(_box(command.data));
      case Pop:
        return _pop();
      case Store:
        return _store(command.data);
      case Load:
        return _load();
      case Execute:
        return _execute();
    }
  }

  void _execute() {
    final fn = _pop();

    if (fn.isNotFunction) {
      throw TypeError();
    }
    final closure = fn as Closure;

    frameStack.add(Frame(parent: closure.parentFrame));
    closure.invoke(this);
    frameStack.removeLast();
  }

  void _load() {
    final key = _pop();
    final value = frameStack.last.get(key);
    _push(value);
  }

  FlightValue _pop() => stack.removeLast();

  void _push(FlightValue value) => stack.add(value);

  FlightValue _box(dynamic value) {
    if (value is FlightValue) return value;
    if (value is bool) return BoolValue(value);
    if (value is num) return NumberValue(value.toDouble());
    if (value is String) return StringValue(value);
    if (value is DartFunction) return DartFunctionValue(value, frameStack.last);
    if (value is Iterable && value.every((it) => it is Command)) {
      return FlightFunctionValue(value as FlightFunction, frameStack.last);
    }
    if (value is Map &&
        value.entries
            .every((it) => it.key is FlightValue && it.value is FlightValue)) {
      return MapValue(value as Map<FlightValue, FlightValue>);
    }
    return DartObjectValue(value);
  }

  void _store(StoreParameters parameters) {
    final key = parameters.key ?? _pop();
    final value = parameters.value ?? _pop();
    final frame = frameStack.last;
    frame.set(key, value, shadows: parameters.shadows);
  }
}
