import 'interfaces.dart';

class BoolValue extends FlightValue<bool> {
  const BoolValue(bool value) : super(value);

  @override
  String get name => "bool";
}

class NumberValue extends FlightValue<double> {
  const NumberValue(double value) : super(value);

  @override
  String get name => "number";
}

class StringValue extends FlightValue<String> {
  const StringValue(String value) : super(value);

  @override
  String get name => "string";
}

/// A closure type can be called, and has a reference to the parent frame
/// at the time of its definition
abstract class Closure {
  Frame get parentFrame;

  void invoke(VirtualMachine vm);
}

/// A Value type that contains a Flight Function (i.e. a List of commands) and
/// implements the Closure type
class FlightFunctionValue extends FlightValue<FlightFunction>
    implements Closure {
  @override
  final Frame parentFrame;

  const FlightFunctionValue(FlightFunction value, this.parentFrame)
      : super(value);

  @override
  String get name => "function";

  @override
  void invoke(VirtualMachine vm) {
    for (final command in value) {
      vm.call(command);
    }
  }
}

/// A Value type that contains a Dart Function and
/// implements the Closure type
class DartFunctionValue extends FlightValue<DartFunction> implements Closure {
  @override
  final Frame parentFrame;

  const DartFunctionValue(DartFunction value, this.parentFrame) : super(value);

  @override
  String get name => "dart-function";

  @override
  void invoke(VirtualMachine vm) => value(vm);
}

class MapValue extends FlightValue<Map<FlightValue, FlightValue>> {
  const MapValue(Map<FlightValue, FlightValue> value) : super(value);

  @override
  String get name => "map";
}

class DartObjectValue extends FlightValue<dynamic> {
  DartObjectValue(value) : super(value);

  @override
  String get name => "object";
}

extension IsFunction on FlightValue {
  bool get isFunction => this is Closure;

  bool get isNotFunction => this is! Closure;
}
